from django.utils.decorators import decorator_from_middleware_with_args

from .middleware import CacheMiddleware


def cache_page(*args, **kwargs):
    cache_alias = kwargs.pop('cache', None)
    return (
        decorator_from_middleware_with_args
        (CacheMiddleware)(cache_alias=cache_alias, *args, **kwargs)
    )



from __future__ import print_function
import os
_this_wo_ext = os.path.basename(__file__).rsplit('.', 1)[0]

__doc__ = """
If you enable `FANCY_REMEMBER_ALL_URLS` then every URL take is turned
into a cache key for cache_page() to remember is recorded.
You can use this to do statistics or to do invalidation by URL.

To use: simply add the URL patterns after like this::

    $ ./manage.py %(this_file)s /path1.html /path3/*/*.json

To show all cached URLs simply run it with no pattern like this::

    $ ./manage.py %(this_file)s

Equally the ``--purge`` switch can always be added. For example,
running this will purge all cached URLs::

    $ ./manage.py %(this_file)s --purge

If you enable `FANCY_REMEMBER_STATS_ALL_URLS` you can get a tally for each
URL how many cache HITS and MISSES it has had.

""" % dict(this_file=_this_wo_ext)

from optparse import make_option

from django.core.management.base import BaseCommand

from fancy_cache.memory import find_urls


class Command(BaseCommand):
    help = __doc__.strip()

    def add_arguments(self, parser):
        parser.add_argument('-p', '--purge', dest='purge', action='store_true',
            help='Purge found URLs')

    args = 'urls'

    def handle(self, *urls, **options):
        verbose = int(options['verbosity']) > 1
        _count = 0
        for url, cache_key, stats in find_urls(urls, purge=options['purge']):
            _count += 1
            if stats:
                self.stdout.write(url[:70].ljust(65)),
                self.stdout.write("HITS", str(stats['hits']).ljust(5)),
                self.stdout.write("MISSES", str(stats['misses']).ljust(5))

            else:
                self.stdout.write(url)

        if verbose:
            self.stdout.write("-- %s URLs cached --" % _count)



import cgi
import functools

from django.core.exceptions import ImproperlyConfigured
from django.conf import settings
from django.core.cache import caches, DEFAULT_CACHE_ALIAS
from django.utils.encoding import iri_to_uri
from django.utils.cache import (
    get_cache_key,
    learn_cache_key,
    patch_response_headers,
    get_max_age
)
from six import string_types, callable
from six.moves.urllib.parse import urlencode

from fancy_cache.utils import md5


REMEMBERED_URLS_KEY = 'fancy-urls'
LONG_TIME = 60 * 60 * 24 * 30


class RequestPath(object):
    def __init__(self, request, only_get_keys, forget_get_keys):
        self.request = request
        self.only_get_keys = only_get_keys
        self.forget_get_keys = forget_get_keys
        assert not (self.only_get_keys and self.forget_get_keys)
        self._prev_get_full_path = request.get_full_path

    def __enter__(self):
        if self.only_get_keys is not None:
            # then monkey patch self.request.get_full_path
            self.request.get_full_path = functools.partial(
                self.get_full_path,
                self.request,
                self.only_get_keys,
                True
            )

        if self.forget_get_keys is not None:
            # then monkey patch self.request.get_full_path
            self.request.get_full_path = functools.partial(
                self.get_full_path,
                self.request,
                self.forget_get_keys,
                False
            )

    def __exit__(self, exc_type, exc_value, traceback):
        if self.only_get_keys is not None or self.forget_get_keys is not None:
            self.request.get_full_path = self._prev_get_full_path

    def get_full_path(self, this, keys, is_only_keys):
        """modified version of django.http.request.Request.get_full_path
        with the ability to return a different query string based on
        `keys` to be included or excluded
        """
        qs = this.META.get('QUERY_STRING', '')
        parsed = cgi.parse_qs(qs, keep_blank_values=True)
        if is_only_keys:
            keep = dict((k, parsed[k]) for k in parsed if k in keys)
        else:
            keep = dict((k, parsed[k]) for k in parsed if k not in keys)
        qs = urlencode(keep, True)
        return '%s%s' % (this.path, ('?' + iri_to_uri(qs)) if qs else '')


class UpdateCacheMiddleware(object):
    """
    Response-phase cache middleware that updates the cache if the response is
    cacheable.

    Must be used as part of the two-part update/fetch cache middleware.
    UpdateCacheMiddleware must be the first piece of middleware in
    MIDDLEWARE_CLASSES so that it'll get called last during the response phase.
    """

    def process_response(self, request, response):
        """Sets the cache, if needed."""
        if (
            not hasattr(request, '_cache_update_cache') or
            not request._cache_update_cache
        ):
            # We don't need to update the cache, just return.
            return response
        if request.method != 'GET':
            # This is a stronger requirement than above. It is needed
            # because of interactions between this middleware and the
            # HTTPMiddleware, which throws the body of a HEAD-request
            # away before this middleware gets a chance to cache it.
            return response
        if not response.status_code == 200:
            return response
        # Try to get the timeout from the "max-age" section of the "Cache-
        # Control" header before reverting to using the default cache_timeout
        # length.
        timeout = get_max_age(response)
        if timeout is None:
            timeout = self.cache_timeout
        elif timeout == 0:
            # max-age was set to 0, don't bother caching.
            return response

        if self.patch_headers:
            patch_response_headers(response, timeout)

        if timeout:
            if callable(self.key_prefix):
                key_prefix = self.key_prefix(request)
            else:
                key_prefix = self.key_prefix
            if self.post_process_response:
                response = self.post_process_response(
                    response,
                    request
                )

            with RequestPath(request, self.only_get_keys, self.forget_get_keys):
                cache_key = learn_cache_key(
                    request,
                    response,
                    timeout,
                    key_prefix
                )

                if self.remember_all_urls:
                    self.remember_url(request, cache_key, timeout)

            self.cache.set(cache_key, response, timeout)

        if self.post_process_response_always:
            response = self.post_process_response_always(
                response,
                request
            )

        return response

    def remember_url(self, request, cache_key, timeout):
        url = request.get_full_path()
        remembered_urls = self.cache.get(REMEMBERED_URLS_KEY, {})
        remembered_urls[url] = cache_key
        self.cache.set(
            REMEMBERED_URLS_KEY,
            remembered_urls,
            LONG_TIME
        )


class FetchFromCacheMiddleware(object):
    """
    Request-phase cache middleware that fetches a page from the cache.

    Must be used as part of the two-part update/fetch cache middleware.
    FetchFromCacheMiddleware must be the last piece of middleware in
    MIDDLEWARE_CLASSES so that it'll get called last during the request phase.
    """
    def process_request(self, request):
        """
        Checks whether the page is already cached and returns the cached
        version if available.
        """
        response = self._process_request(request)
        if self.remember_stats_all_urls:
            # then we're nosy
            cache_key = request.get_full_path()
            if response is None:
                cache_key += '__misses'
            else:
                cache_key += '__hits'
            cache_key = md5(cache_key)
            if self.cache.get(cache_key) is None:
                self.cache.set(cache_key, 0, LONG_TIME)
            self.cache.incr(cache_key)
        return response

    def _process_request(self, request):
        if self.cache_anonymous_only:
            if not hasattr(request, 'user'):
                raise ImproperlyConfigured(
                    "The Django cache middleware with "
                    "CACHE_MIDDLEWARE_ANONYMOUS_ONLY=True requires "
                    "authentication middleware to be installed. Edit your "
                    "MIDDLEWARE_CLASSES setting to insert "
                    "'django.contrib.auth.middleware.AuthenticationMiddleware'"
                    "before the CacheMiddleware."
                )

        if request.method not in ('GET', 'HEAD'):
            request._cache_update_cache = False
            # Don't bother checking the cache.
            return None

        #if (
        #    request.GET and
        #    not callable(self.key_prefix) and
        #    not self.only_get_keys
        #):
        #    request._cache_update_cache = False
        #    # Default behaviour for requests with GET parameters: don't bother
        #    # checking the cache.
        #    return None

        if self.cache_anonymous_only and request.user.is_authenticated():
            request._cache_update_cache = False
            # Don't cache requests from authenticated users.
            return None

        if callable(self.key_prefix):
            key_prefix = self.key_prefix(request)
            if key_prefix is None:
                request._cache_update_cache = False
                # Don't bother checking the cache if key_prefix function
                # returns magic "None" value.
                return None
        else:
            key_prefix = self.key_prefix

        with RequestPath(request, self.only_get_keys, self.forget_get_keys):
            cache_key = get_cache_key(request, key_prefix)

        if cache_key is None:
            request._cache_update_cache = True
            # No cache information available, need to rebuild.
            return None

        response = self.cache.get(cache_key, None)
        if response is None:
            request._cache_update_cache = True
            # No cache information available, need to rebuild.
            return None

        request._cache_update_cache = False
        if self.post_process_response_always:
            response = self.post_process_response_always(
                response,
                request=request
            )

        return response


class CacheMiddleware(UpdateCacheMiddleware, FetchFromCacheMiddleware):
    """
    Cache middleware that provides basic behavior for many simple sites.

    Also used as the hook point for the cache decorator, which is generated
    using the decorator-from-middleware utility.

    :param cache_timeout:
        How many seconds to cache the page. This is ignored if the view sets a
        Cache-Control header with a max-age.

    :param key_prefix:
        Either a string or a callable function. If it's a callable function,
        it's called with the request as the first and only argument.

    :param cache_anonymous_only:
        Guess!

    :param patch_headers:
        Basically, if you set a cache_timeout of 60 it additionally sets a
        Expires header with that timeout.

    :param post_process_response:
        Callable function that gets called with the response (and request) just
        before the response gets set in cache.

    :param post_process_response_always:
        Callable function that gets called with the response (and request)
        every time the response goes through the middleware cached or not.

    :param only_get_keys:
        List of query string keys to reduce the cache key to. Without this a
        GET /some/path?foo=bar and /some/path?foo=bar&other=junk gets two
        different cache keys when it could be that the `other=junk` parameter
        doesn't change anything.

    :param forget_get_keys:
        List of query string keys to ignore when reducing the cache key.

    :param remember_all_urls:
        With this option you can have all cached URLs stored in cache which
        can make it easy to do things like cache invalidation by URL.

    :param remember_stats_all_urls:
        Only applicable if `remember_all_urls` is set. This stores a count
        of the number of times a `cache_page` hits and misses.

    """
    def __init__(
        self,
        cache_timeout=settings.CACHE_MIDDLEWARE_SECONDS,
        key_prefix=settings.CACHE_MIDDLEWARE_KEY_PREFIX,
        cache_anonymous_only=getattr(
            settings,
            'CACHE_MIDDLEWARE_ANONYMOUS_ONLY',
            False
        ),
        patch_headers=False,
        post_process_response=None,
        post_process_response_always=None,
        only_get_keys=None,
        forget_get_keys=None,
        remember_all_urls=getattr(
            settings,
            'FANCY_REMEMBER_ALL_URLS',
            False
        ),
        remember_stats_all_urls=getattr(
            settings,
            'FANCY_REMEMBER_STATS_ALL_URLS',
            False
        ),
        **kwargs
    ):
        self.patch_headers = patch_headers
        self.cache_timeout = cache_timeout
        self.key_prefix = key_prefix
        self.cache_anonymous_only = cache_anonymous_only
        self.post_process_response = post_process_response
        self.post_process_response_always = post_process_response_always
        if isinstance(only_get_keys, string_types):
            only_get_keys = [only_get_keys]
        self.only_get_keys = only_get_keys
        if isinstance(forget_get_keys, string_types):
            forget_get_keys = [forget_get_keys]
        self.forget_get_keys = forget_get_keys
        self.remember_all_urls = remember_all_urls
        self.remember_stats_all_urls = remember_stats_all_urls

        try:
            cache_alias = kwargs['cache_alias']
            if cache_alias is None:
                cache_alias = DEFAULT_CACHE_ALIAS
        except KeyError:
            cache_alias = settings.CACHE_MIDDLEWARE_ALIAS

        self.cache_alias = cache_alias
        self.cache = caches[self.cache_alias]



from django.shortcuts import render
from django.conf import settings

from fancy_cache.memory import find_urls


def home(request):
    data = {
        'found': find_urls([]),
        'remember_all_urls_setting': getattr(
            settings,
            'FANCY_REMEMBER_ALL_URLS',
            False
        ),
        'remember_stats_all_urls_setting': getattr(
            settings,
            'FANCY_REMEMBER_STATS_ALL_URLS',
            False
        )
    }
    return render(request, 'fancy-cache/home.html', data)



from django.conf.urls.defaults import url
from . import views


urlpatterns = [
    url(r'^$', views.home, name='home'),
    #url(r'^2$', views.home2, name='home2'),
    #url(r'^3$', views.home3, name='home3'),
]



from django.test import TestCase
from nose.tools import eq_, ok_
from django.core.cache import cache
from django.core.management import call_command
from django.test import TestCase
from django.utils.six import StringIO
from fancy_cache.middleware import REMEMBERED_URLS_KEY


class TestBaseCommand(TestCase):
    def setUp(self):
        self.urls = {
            '/page1.html': 'key1',
            '/page2.html': 'key2',
            '/page3.html?foo=bar': 'key3',
            '/page3.html?foo=else': 'key4',
        }
        for key, value in self.urls.items():
            cache.set(value, key)
        cache.set(REMEMBERED_URLS_KEY, self.urls, 5)

    def tearDown(self):
        cache.clear()

    def test_fancyurls_command(self):
        out = StringIO()
        call_command('fancy-urls', verbosity=3, stdout=out)
        self.assertIn('4 URLs cached', out.getvalue())

    def test_purge_command(self):
        out = StringIO()
        # Note: first call will show 4 URLs, so call again to confirm deletion
        call_command('fancy-urls', '--purge')
        call_command('fancy-urls', verbosity=3, stdout=out)
        self.assertIn('0 URLs cached', out.getvalue())



import os
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

SECRET_KEY = 'anything'

ALLOWED_HOSTS = ('testserver',)

DATABASES = {
    'default': {
        'NAME': ':memory:',
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

CACHES = {
    'default': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake'
    },
    'second_backend': {
        'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        'LOCATION': 'unique-snowflake'
    }
}

INSTALLED_APPS = [
    'fancy_cache',
    'fancy_tests.tests',

    'django.contrib.auth',
    'django.contrib.contenttypes',
]


ROOT_URLCONF = 'fancy_tests.tests.urls'


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]



import unittest

from nose.tools import eq_, ok_
from django.core.cache import cache

from fancy_cache.middleware import REMEMBERED_URLS_KEY
from fancy_cache.memory import find_urls


class TestMemory(unittest.TestCase):

    def setUp(self):
        self.urls = {
            '/page1.html': 'key1',
            '/page2.html': 'key2',
            '/page3.html?foo=bar': 'key3',
            '/page3.html?foo=else': 'key4',
        }
        for key, value in self.urls.items():
            cache.set(value, key)
        cache.set(REMEMBERED_URLS_KEY, self.urls, 5)

    def tearDown(self):
        cache.clear()

    def test_find_all_urls(self):
        found = list(find_urls([]))
        eq_(len(found), 4)
        for key, value in self.urls.items():
            pair = (key, value, None)
            ok_(pair in found)

    def test_find_and_purge_all_urls(self):
        found = list(find_urls([], purge=True))
        eq_(len(found), 4)
        for key, value in self.urls.items():
            pair = (key, value, None)
            ok_(pair in found)
        found = list(find_urls([]))
        eq_(len(found), 0)

    def test_find_one_url(self):
        found = list(find_urls(['/page1.html']))
        eq_(len(found), 1)
        ok_(('/page1.html', 'key1', None) in found)

    def test_purge_one_url(self):
        ok_(cache.get('key1'))
        ok_('/page1.html' in cache.get(REMEMBERED_URLS_KEY))
        found = list(find_urls(['/page1.html'], purge=True))
        eq_(len(found), 1)
        ok_(('/page1.html', 'key1', None) in found)

        ok_(not cache.get('key1'))
        ok_('/page1.html' not in cache.get(REMEMBERED_URLS_KEY))
        # find all the rest in there
        found = list(find_urls([]))
        eq_(len(found), 3)
        ok_(('/page1.html', 'key1', None) not in found)

    def test_some_urls(self):
        found = list(find_urls(['/page2.html*']))
        eq_(len(found), 1)
        ok_(('/page2.html', 'key2', None) in found)

    def test_some_urls_double_star(self):
        found = list(find_urls(['/page*.html?*']))
        eq_(len(found), 2)
        ok_(('/page3.html?foo=bar', 'key3', None) in found)
        ok_(('/page3.html?foo=else', 'key4', None) in found)



import uuid
from django.shortcuts import render
from fancy_cache import cache_page


def _view(request):
    random_string = uuid.uuid4().hex
    return render(request, 'home.html', dict(random_string=random_string))


@cache_page(60)
def home(request):
    return _view(request)


def prefixer1(request):
    if request.META.get('AUTH_USER'):
        # disable
        return None
    return 'a_key'


@cache_page(60, key_prefix=prefixer1)
def home2(request):
    return _view(request)


def post_processor1(response, request):
    assert 'In your HTML' not in response.content.decode("utf8")
    response.content += ('In your HTML:%s' % uuid.uuid4().hex).encode("utf8")
    return response


@cache_page(60,
            key_prefix=prefixer1,
            post_process_response=post_processor1)
def home3(request):
    return _view(request)


@cache_page(60,
            key_prefix=prefixer1,
            post_process_response_always=post_processor1)
def home4(request):
    return _view(request)


@cache_page(60, only_get_keys=['foo', 'bar'])
def home5(request):
    return _view(request)

@cache_page(60, forget_get_keys=['bar'])
def home5bis(request):
    return _view(request)

@cache_page(60, remember_stats_all_urls=True, remember_all_urls=True)
def home6(request):
    return _view(request)

@cache_page(60, cache='second_backend')
def home7(request):
    return _view(request)



import codecs
import os
import re


# Prevent spurious errors during `python setup.py test`, a la
# http://www.eby-sarna.com/pipermail/peak/2010-May/003357.html:
try:
    import multiprocessing
except ImportError:
    pass

from setuptools import setup, find_packages


def read(*parts):
    content = codecs.open(os.path.join(os.path.dirname(__file__), *parts)).read()
    try:
        return content.decode("utf8")
    except AttributeError:
        # python 3
        return content


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='django-fancy-cache',
    version=find_version('fancy_cache/__init__.py'),
    description="A Django 'cache_page' decorator on steroids",
    long_description=read('README.rst'),
    author='Peter Bengtsson',
    author_email='mail@peterbe.com',
    license='BSD',
    install_requires=['six'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    tests_require=['nose'],
    test_suite='runtests.runtests',
    url='https://github.com/peterbe/django-fancy-cache'
)



import re

from django.core.cache import cache

from fancy_cache.middleware import REMEMBERED_URLS_KEY, LONG_TIME
from fancy_cache.utils import md5

__all__ = ('find_urls',)


def _match(url, regexes):
    if not regexes:
        return url
    for regex in regexes:
        if regex.match(url):
            return True
    return False


def _urls_to_regexes(urls):
    regexes = []
    for each in urls:
        parts = each.split('*')
        if len(parts) == 1:
            regexes.append(re.compile('^%s$' % re.escape(parts[0])))
        else:
            _re = '.*'.join(re.escape(x) for x in parts)
            regexes.append(re.compile('^%s$' % _re))
    return regexes


def find_urls(urls=None, purge=False):
    remembered_urls = cache.get(REMEMBERED_URLS_KEY, {})
    _del_keys = []
    if urls:
        regexes = _urls_to_regexes(urls)
    for url in remembered_urls:
        if not urls or _match(url, regexes):
            cache_key = remembered_urls[url]
            if not cache.get(cache_key):
                continue
            if purge:
                cache.delete(cache_key)
                _del_keys.append(url)
            misses_cache_key = '%s__misses' % url
            misses_cache_key = md5(misses_cache_key)
            hits_cache_key = '%s__hits' % url
            hits_cache_key = md5(hits_cache_key)

            misses = cache.get(misses_cache_key)
            hits = cache.get(hits_cache_key)
            if misses is None and hits is None:
                stats = None
            else:
                stats = {
                    'hits': hits or 0,
                    'misses': misses or 0
                }
            yield (url, cache_key, stats)

    if _del_keys:
        # means something was changed
        for url in _del_keys:
            remembered_urls.pop(url)
            misses_cache_key = '%s__misses' % url
            hits_cache_key = '%s__hits' % url
            cache.delete(misses_cache_key)
            cache.delete(hits_cache_key)

        cache.set(
            REMEMBERED_URLS_KEY,
            remembered_urls,
            LONG_TIME
        )



import unittest
import re
from nose.tools import eq_, ok_
from django.test.client import RequestFactory
from django.core.cache import cache, caches
from fancy_cache.memory import find_urls

from . import views


class TestViews(unittest.TestCase):

    def setUp(self):
        self.factory = RequestFactory()

    def tearDown(self):
        cache.clear()

    def test_render_home1(self):
        request = self.factory.get('/anything')

        response = views.home(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]

        # do it again
        response = views.home(request)
        eq_(response.status_code, 200)
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        eq_(random_string_1, random_string_2)

    def test_render_home2(self):
        authenticated = RequestFactory(AUTH_USER='peter')
        request = self.factory.get('/2')

        response = views.home2(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]

        # do it again
        response = views.home2(request)
        eq_(response.status_code, 200)
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        eq_(random_string_1, random_string_2)

        # do it again, but with a hint to disable cache
        request = authenticated.get('/2')
        response = views.home2(request)
        eq_(response.status_code, 200)
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_(random_string_1 != random_string_2)

    def test_render_home3(self):
        request = self.factory.get('/anything')

        response = views.home3(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_('In your HTML' in response.content.decode("utf8"))
        extra_random_1 = re.findall('In your HTML:(\w+)', response.content.decode("utf8"))[0]

        response = views.home3(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        extra_random_2 = re.findall('In your HTML:(\w+)', response.content.decode("utf8"))[0]
        ok_('In your HTML' in response.content.decode("utf8"))
        eq_(random_string_1, random_string_2)
        # the post_process_response is only called once
        eq_(extra_random_1, extra_random_2)

    def test_render_home3_no_cache(self):
        factory = RequestFactory(AUTH_USER='peter')
        request = factory.get('/3')

        response = views.home3(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        ok_('In your HTML' not in response.content.decode("utf8"))

    def test_render_home4(self):
        request = self.factory.get('/4')

        response = views.home4(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_('In your HTML' in response.content.decode("utf8"))
        extra_random_1 = re.findall('In your HTML:(\w+)', response.content.decode("utf8"))[0]

        response = views.home4(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        extra_random_2 = re.findall('In your HTML:(\w+)', response.content.decode("utf8"))[0]
        ok_('In your HTML' in response.content.decode("utf8"))
        eq_(random_string_1, random_string_2)
        # the post_process_response is now called every time
        ok_(extra_random_1 != extra_random_2)

    def test_render_home5(self):
        request = self.factory.get('/4', {'foo': 'bar'})
        response = views.home5(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]

        request = self.factory.get('/4', {'foo': 'baz'})
        response = views.home5(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_(random_string_1 != random_string_2)

        request = self.factory.get('/4', {'foo': 'baz', 'other': 'junk'})
        response = views.home5(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_3 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        eq_(random_string_2, random_string_3)

    def test_render_home5bis(self):
        request = self.factory.get('/4', {'foo': 'bar'})
        response = views.home5bis(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]

        request = self.factory.get('/4', {'foo': 'baz'})
        response = views.home5bis(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_(random_string_1 != random_string_2)

        request = self.factory.get('/4', {'foo': 'baz', 'bar': 'foo'})
        response = views.home5bis(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_3 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        eq_(random_string_2, random_string_3)

    def test_remember_stats_all_urls(self):
        request = self.factory.get('/anything')
        response = views.home6(request)
        eq_(response.status_code, 200)

        # now ask the memory thing
        match, = find_urls(urls=['/anything'])
        eq_(match[0], '/anything')
        eq_(match[2]['hits'], 0)
        eq_(match[2]['misses'], 1)

        # second time
        response = views.home6(request)
        eq_(response.status_code, 200)
        match, = find_urls(urls=['/anything'])
        eq_(match[0], '/anything')
        eq_(match[2]['hits'], 1)
        eq_(match[2]['misses'], 1)

    def test_remember_stats_all_urls_looong_url(self):
        request = self.factory.get(
            '/something/really/long/to/start/with/right/here/since/this/will/'
            'test/that/things/work/with/long/urls/too',
            {
                'line1': 'Bad luck, wind been blowing at my back',
                'line2': "I was born to bring trouble to wherever I'm at",
                'line3': "Got the number thirteen, tattooed on my neck",
                'line4': "When the ink starts to itch, ",
                'line5': "then the black will turn to red",
            }
        )
        response = views.home6(request)
        eq_(response.status_code, 200)

        # now ask the memory thing
        match, = find_urls()
        ok_(match[0].startswith('/something/really'))
        eq_(match[2]['hits'], 0)
        eq_(match[2]['misses'], 1)

        # second time
        response = views.home6(request)
        eq_(response.status_code, 200)
        match, = find_urls([])
        ok_(match[0].startswith('/something/really'))
        eq_(match[2]['hits'], 1)
        eq_(match[2]['misses'], 1)

    def test_cache_backends(self):
        request = self.factory.get('/anything')

        response = views.home7(request)
        eq_(response.status_code, 200)
        ok_(re.findall('Random:\w+', response.content.decode("utf8")))
        random_string_1 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]

        # clear second cache backend
        caches['second_backend'].clear()
        response = views.home7(request)
        eq_(response.status_code, 200)
        random_string_2 = re.findall('Random:(\w+)', response.content.decode("utf8"))[0]
        ok_(random_string_1 != random_string_2)
